#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import wx

from Table    import Table
from Relation import Relation

# ----------------------------------------------------------------------
# Operationen auf einer *.ERMzilla-Datei.
#
# gültige Keywords innerhalb der Datei:
#   #                                                                   - Kommentar
#   TABLE        table_name                                             - Start und Benennung einer Tabelle
#   "  "         column_name colum_type column_attributes               - Definition einer Spalte
#   TABEND                                                              - kennzeichnet das Ende einer inline-Tabelle
#   TABPOS       table_name (x,y) {(x,y)}                               - optional (Positionsangaben)
#   RELATION     rel_name rel_dir tab1.col tab2.col [border1 border2]   - Definition einer Relation
#   RELPOS       rel_name (x,y) {(x,y)}                                 - optional (Positionsangaben)
#   IMPORT_TABLE file_name                                              - Tabellen-Definition aus Datei lesen (mit Pfad relativ zur geladenen Datei)
class FileOps():
    def __init__(self, tables, relations):
        self.tables=tables
        self.relations=relations
        self.filename=None
        self.dirname=None
        self.relation_name_generator_last=0
        self.log=[]

    # ------------------------------------------------------------------
    # Datei laden.
    def load(self, filename):
        if not os.path.isfile(filename):
            self.log.append("Error! File not found")
            return
        self.filename=filename
        self.dirname=os.path.dirname(filename)
        if self.__loadTables(self.filename):                            # ohne Tabellen brauchts auch keine Relationen
            self.__loadRelations(self.filename)

    # ------------------------------------------------------------------
    #
    def changedOnDisk(self):
        pass

    # ------------------------------------------------------------------
    #
    def saveChanges(self):
        pass

    # ------------------------------------------------------------------
    # Liest die Tabellen-Definitionen aus der Datei filename.
    # Beim Keyword IMPORT_TABLE ruft sich diese Funtion selbst auf.
    # In diesem Fall ist imported_filename zu übergeben, damit keine
    # Endlos-Rekurstion entstehen kann.
    def __loadTables(self, filename, imported_filename=None):
        with open(filename, "rt") as fl:
            ln_cnt=0
            table_name=None
            cols=[]
            for ln in fl:
                ln_cnt+=1
                if ln.strip()=="" or ln.startswith("#"):                # Leerzeilen und Kommentare überlesen
                    continue
                lns=ln.split()
                if ln.startswith("TABLE "):
                    if table_name:
                        t_tmp=Table(table_name, cols, (0,0), imported_filename)
                        self.tables.append(t_tmp)
                        table_name=None
                        cols=[]
                    if len(lns)==2:
                        table_name=lns[1]
                    elif len(lns)>2:
                        table_name=lns[1]
                        self.log.append("Warning! Too many Elements in line %d"%(ln_cnt,))
                    else:
                        self.log.append("Error! table_name missing in line %d"%(ln_cnt,))
                        return False
                    if self.__getTable(table_name):
                        self.log.append("Error! duplicate table declaration in line %d"%(ln_cnt,))
                        return False
                elif ln.startswith("  "):
                    if not table_name:
                        self.log.append("Warning! Column outside Table in line %d"%(ln_cnt,))
                        continue
                    if len(lns)==3:
                        cols.append(tuple(lns))
                    elif len(lns)==2:
                        cols.append((lns[0], lns[1], ""))
                    elif len(lns)>3:
                        cols.append((lns[0], lns[1], lns[2]))
                        self.log.append("Warning! Too many Elements in line %d"%(ln_cnt,))
                    else:
                        self.log.append("Error! Attribute[s] missing in line %d"%(ln_cnt,))
                        return False
                elif ln.strip()=="TABEND":
                    if not table_name:
                        self.log.append("Warning! TABEND outside Table in line %d"%(ln_cnt,))
                        continue
                    t_tmp=Table(table_name, cols, (0,0), imported_filename)
                    self.tables.append(t_tmp)
                    table_name=None
                    cols=[]
                elif ln.startswith("TABPOS ") and not imported_filename:
                    if table_name:
                        t_tmp=Table(table_name, cols, (0,0))
                        self.tables.append(t_tmp)
                        table_name=None
                        cols=[]
                    if len(lns)<3:
                        self.log.append("Error! Attribute[s] missing in line %d"%(ln_cnt,))
                        return False
                    elif len(lns)>3:
                        self.log.append("Warning! Too many Elements in line %d"%(ln_cnt,))
                    tn=self.__getTable(lns[1])
                    if not tn:
                        self.log.append("Error! Unknown table_name in line %d"%(ln_cnt,))
                        return False
                    pos=self.__parseTuple(lns[2])
                    if pos:
                        tn.pos=pos
                    else:
                        self.log.append("Error! Invalid coordinate in line %d"%(ln_cnt,))
                        return False
                elif ln.startswith("IMPORT_TABLE ") and not imported_filename:
                    if table_name:
                        t_tmp=Table(table_name, cols, (0,0))
                        self.tables.append(t_tmp)
                        table_name=None
                        cols=[]
                    if len(lns)>=2:
                        if len(lns)>2:
                            self.log.append("Warning! Too many Elements in line %d"%(ln_cnt,))
                        fn=lns[1]
                        if not os.path.isfile(fn):
                            fn=os.path.join(self.dirname, lns[1])
                        if os.path.isfile(fn):
                            if not self.__loadTables(fn, imported_filename=lns[1]):
                                self.log.append("Info! last error was in line %d"%(ln_cnt,))
                                return False
                        else:
                            self.log.append("Error! File not found in line %d / %s"%(ln_cnt, lns[1]))
                            return False
                    else:
                        self.log.append("Error! Filename missing in line %d"%(ln_cnt,))
                        return False
        if table_name:
            t_tmp=Table(table_name, cols, (0,0), imported_filename)
            self.tables.append(t_tmp)
        return True

    # ------------------------------------------------------------------
    # Liest die Relations-Definitionen aus der Datei filename.
    def __loadRelations(self, filename):
        with open(filename, "rt") as fl:
            ln_cnt=0
            relation_names=set()
            for ln in fl:
                ln_cnt+=1
                if ln.strip()=="" or ln.startswith("#"):                # Leerzeilen und Kommentare überlesen
                    continue
                lns=ln.split()
                if ln.startswith("RELATION "):
                    if len(lns)<5:
                        self.log.append("Error! Attribute[s] missing in line %d"%(ln_cnt,))
                        return False
                    elif len(lns)>7:
                        self.log.append("Warning! Too many Elements in line %d"%(ln_cnt,))
                    if len(lns)==7:
                        _, rel_name, rel_dir, tab1_col, tab2_col, border1, border2=lns
                    elif len(lns)==6:
                        _, rel_name, rel_dir, tab1_col, tab2_col, border1=lns
                        border2="R"
                    elif len(lns)==5:
                        _, rel_name, rel_dir, tab1_col, tab2_col=lns
                        border1=border2="R"
                    else:
                        _, rel_name, rel_dir, tab1_col, tab2_col, border1, border2=lns[:7]
                    if rel_name=="*":
                        rel_name=self.__relation_name_generator()
                    if len(rel_name)<3:
                        self.log.append("Warning! rel_name too short in line %d"%(ln_cnt,))
                    #if len(rel_dir)!=3:# or rel_dir not in ("1-1", "1-n", "n-1", "m-n", "n-m"):
                    #    self.log.append("Error! invalid rel_dir in line %d"%(ln_cnt,))
                    #    return False
                    if not rel_dir.find("-")>0:
                        self.log.append("Error! invalid rel_dir in line %d"%(ln_cnt,))
                        return False
                    if not self.__isValidTabCol(tab1_col):
                        self.log.append("Error! invalid tab1.col in line %d"%(ln_cnt,))
                        return False
                    if not self.__isValidTabCol(tab2_col):
                        self.log.append("Error! invalid tab2.col in line %d"%(ln_cnt,))
                        return False
                    if border1 not in ("L", "R"):
                        self.log.append("Error! invalid border1 in line %d"%(ln_cnt,))
                        return False
                    if border2 not in ("L", "R"):
                        self.log.append("Error! invalid border2 in line %d"%(ln_cnt,))
                        return False
                    if rel_name in relation_names:
                        self.log.append("Error! duplicate rel_name in line %d"%(ln_cnt,))
                        return False
                    relation_names.add(rel_name)
                    r_tmp=Relation(self.tables, rel_name, rel_dir, tab1_col, tab2_col, border1, border2)
                    self.relations.append(r_tmp)
                elif ln.startswith("RELPOS "):
                    if len(lns)<3:
                        self.log.append("Error! Attribute[s] missing in line %d"%(ln_cnt,))
                        return False
                    if lns[1] not in relation_names:
                        self.log.append("Error! Unknown rel_name in line %d"%(ln_cnt,))
                        return False
                    pl=[]
                    for p in lns[2:]:
                        pt=self.__parseTuple(p)
                        if not pt:
                            self.log.append("Error! Invalid coordinate in line %d"%(ln_cnt,))
                            return False
                        pl.append(pt)
                    r=self.__getRelation(lns[1])
                    if r:
                        r.line=pl
                    else:
                        # kann eigentlich nicht vorkommen....
                        self.log.append("Error! rel_name not found in line %d"%(ln_cnt,))
                        return False

    # ------------------------------------------------------------------
    # Liefert das Tabellen-Objekt mit dem Namen table_name oder None.
    def __getTable(self, table_name):
        for t in self.tables:
            if t.name==table_name:
                return t
        return None

    # ------------------------------------------------------------------
    # Liefert das Relations-Objekt mit den Namen relation_name oder
    # None.
    def __getRelation(self, relation_name):
        for r in self.relations:
            if r.name==relation_name:
                return r
        return None

    # ------------------------------------------------------------------
    # Liefert True, wenn tab_col bereits bekannt ist.
    def __isValidTabCol(self, tab_col):
        for t in self.tables:
            for c, _, _ in t.cols:
                if "%s.%s"%(t.name, c)==tab_col:
                    return True
        return False

    # ------------------------------------------------------------------
    # Zerlegt einen String in strg im Format "(111,222)" und liefert es
    # als Tupel (111, 222) zurück. Im Fehlerfall wird None geliefert.
    def __parseTuple(self, strg):
        if strg.startswith("(") and strg.endswith(")"):
            ss=strg.strip("(").strip(")").split(",")
            if len(ss)==2:
                try:
                    a=int(ss[0])
                    b=int(ss[1])
                    return (a, b)
                except ValueError:
                    return None
            else:
                return None
        else:
            return None

    # ------------------------------------------------------------------
    # Liefert einen (neuen) eindeutigen Relationsnamen.
    def __relation_name_generator(self):
        self.relation_name_generator_last+=1
        return "rel_%06d"%(self.relation_name_generator_last)

